import traceback
import pathlib
import os


def dataset_upload(collection_id,  upload_path="."):
    from .baai_environment import setup_network

    setup_network(None)

    upload_path = pathlib.Path(upload_path)

    try:
        from .baai_uploader import data_uploader

        data_uploader(collection_id, upload_path)
    except Exception: # noqa
        traceback.print_exc()
    except KeyboardInterrupt:
        pass



import base64
import traceback

import jwt
import msgpack

# 更简洁的版本，直接返回原始数据
def jwtsign_read_extract(download_sign):
    try:
        # 解码JWT而不验证签名
        payload = jwt.decode(
            download_sign.strip(),
            options={"verify_signature": False}  # 不验证签名
        )
        download_data = payload["download_path"]
        download_pack = base64.b64decode(download_data)
        download_info = msgpack.loads(download_pack, raw=False)

        return {
            "proto": download_info["proto"],
            "prefix": download_info["prefix"],
            "path": download_info["path"],
            "check_point_hash": "127.0.0.1"
        }

    except Exception as e:
        traceback.print_exc()
        print(f"Decode error: {e}")
        return None


import json
import base64

from pydantic import BaseModel

from .baai_config import Application


class JwtClaims(BaseModel):
    sub : str
    iat : int
    exp : int


def claims_parse(jwt_token):
    parts = jwt_token.split('.')
    if len(parts) != 3:
        return None
    payload = parts[1]

    padding = 4 - len(payload) % 4
    if padding != 4:
        payload += '=' * padding

    decoded_bytes = base64.b64decode(payload)
    decoded_str = decoded_bytes.decode('utf-8')
    return json.loads(decoded_str)




from pydantic import BaseModel, Field


class ReqSignCreate(BaseModel):
    paths: list[str] = Field(default_factory=list)

class RespSignCreage(BaseModel):
    path: str
    endpoint: str


class JSONResponse(BaseModel):
    code:  int
    message: str
    data: RespSignCreage


class MetaData(BaseModel):
    video_id: int = Field(alias="videoId")
    os_path: str = Field(alias="osPath")

    def get_storage_path(self) -> str:
        return self.os_path



import hashlib
import pathlib
from concurrent.futures import ThreadPoolExecutor
from concurrent.futures import as_completed

import requests


from .baai_config import Application, Progress
from .baai_downloder import check_file
from .core.baai_presign import read_presign

app = Application() # noqa
progress = Progress()

executor_spawn = ThreadPoolExecutor(max_workers=10)
executor_progress = ThreadPoolExecutor(max_workers=app.jobs)
executor_sign = ThreadPoolExecutor(max_workers=app.jobs)
executor_down = ThreadPoolExecutor(max_workers=app.jobs)
executor_watch = ThreadPoolExecutor(max_workers=50)
executor_merge = ThreadPoolExecutor(max_workers=app.jobs/2)


def download_sign_file(raw_data: str):
    from .presign import extract


    config = Application()

    meta_file: dict[str, str] = extract.jwtsign_read_extract(raw_data)
    hash_prefix = hashlib.md5(meta_file.get("path").encode("utf-8")).hexdigest()
    hash_ext = pathlib.Path(meta_file.get("path")).suffix

    sign_addr: str = read_presign(raw_data)
    resp_meta = requests.head(sign_addr, headers={"accept-encoding": None}) # noqa
    size = int(resp_meta.headers.get('Content-Length'))

    progress.add_require_size(size)
    # 文件存在不进行下载
    if check_file(f"{config.save_path}/{hash_prefix}{hash_ext}", size):
        progress.add_download_size(size)
        progress.add_download_count(1)
        return

    chunk_size = config.chunk_size
    total_parts = int((size + chunk_size - 1) / chunk_size)

    save_path = pathlib.Path(config.save_path)
    save_path.mkdir(parents=True, exist_ok=True)

    download_futures = []
    for part_index in range(total_parts):
        start = part_index * chunk_size
        end = min(start + chunk_size, size)
        task_down = executor_down.submit(download_sign_part_write, raw_data, start, end, hash_prefix)
        download_futures.append(task_down)

    # 等待完成
    for future in as_completed(download_futures):
        future.result()

    executor_merge.submit(download_sign_part_merge,  hash_prefix, hash_ext, total_parts)



def download_sign_part_write(raw_data, start: int, end: int, part_hash):
    config = Application()

    part_idx = int(start / config.chunk_size)
    part_name =  pathlib.Path(f"{config.save_path}/{part_hash}_{part_idx}.bin")

    if check_file(part_name.__str__(),  end-start):
        pass
    else:
        range_header = f"bytes={start}-{end - 1}"
        sign_addr = read_presign(raw_data)
        resp_file = requests.get(sign_addr, headers={"Range": range_header})

        # md5加密
        with part_name.open("wb") as fwriter:
            fwriter.write(resp_file.content)

    progress.add_download_size(end-start)


def download_sign_part_merge(name_prefix, name_ext, total_parts): # noqa

    config = Application()
    target_path = pathlib.Path(f"{config.save_path}/{name_prefix}{name_ext}")
    with target_path.open("ab") as fwriter:
        for part_idx in range(total_parts):
            part_name = pathlib.Path(f"{config.save_path}/{name_prefix}_{part_idx}.bin")
            with part_name.open("rb") as f:
                data = f.read()
                fwriter.write(data)

    for part_idx in range(total_parts):
        part_name = pathlib.Path(f"{config.save_path}/{name_prefix}_{part_idx}.bin")
        part_name.unlink()
    progress.add_download_count(1)


def download_progress():
    config = Application()
    with open(config.mate_path, "r") as fr:
        for _line in fr:
            try:
                executor_progress.submit(download_sign_file, _line)
            except Exception as e: # noqa
                progress.add_download_count(-1)


def download_sign_dataset(dataset_id, save_path, meta_path, jobs=8):
    import time
    from tqdm import tqdm # noqa

    from .baai_config import Application, Progress # noqa
    from .baai_logger import logger_download

    progress = Progress()
    logger_download()

    from .baai_downloder import executor_spawn, download_readrows, download_watch


    p1 = executor_spawn.submit(download_readrows)
    p2 = executor_spawn.submit(download_progress)
    p3 = executor_spawn.submit(download_watch)

    with tqdm(
        desc="speed",
        # colour="green",
        ncols=100,
        ascii="░▒▓█",
        unit="B",
        unit_scale=True,
        # bar_format="{desc}: {percentage:3.1f}%|{bar}| {n_fmt}/{total_fmt} [{elapsed}<{remaining}, {rate_fmt}] {postfix}",
        bar_format="[{elapsed}] | {percentage:3.1f}%|{bar}| {n_fmt}/{total_fmt}{postfix} [{desc}: {rate_fmt}]"
    ) as pbar:
        last_size = 0
        while True:
            time.sleep(1)
            # 更新总任务数
            pbar.total = progress.require_size
            pbar.refresh()  # 刷新以显示新的total

            # 计算增量并更新进度条
            current_size = progress.download_size
            increment = min(current_size, progress.require_size) - last_size

            if increment > 0:
                pbar.update(increment)
                last_size = current_size

            pbar.set_postfix({
                "count": f"{progress.download_count}/{progress.require_count}",
            })

            if progress.require_size == progress.download_size and progress.require_count == progress.download_count:
                break

    p1.result()
    p2.result()
    p3.result()

    logger_download()


from concurrent.futures import ThreadPoolExecutor as BaseThreadPoolExecutor, _base
import threading
import weakref
import queue


_threads_queues = weakref.WeakKeyDictionary()
_shutdown = False


def _worker(executor_reference, work_queue, initializer, initargs):
    if initializer is not None:
        try:
            initializer(*initargs)
        except BaseException:
            _base.LOGGER.critical('Exception in initializer:', exc_info=True)
            executor = executor_reference()
            if executor is not None:
                executor._initializer_failed()
            return
    try:
        while True:
            try:
                work_item = work_queue.get_nowait()
            except queue.Empty:
                # attempt to increment idle count if queue is empty
                executor = executor_reference()
                if executor is not None:
                    executor._idle_semaphore.release()
                del executor
                work_item = work_queue.get(block=True)

            if work_item is not None:
                work_item.run()
                # Delete references to object. See GH-60488
                del work_item
                continue

            executor = executor_reference()
            # Exit if:
            #   - The interpreter is shutting down OR
            #   - The executor that owns the worker has been collected OR
            #   - The executor that owns the worker has been shutdown.
            if _shutdown or executor is None or executor._shutdown:
                # Flag the executor as shutting down as early as possible if it
                # is not gc-ed yet.
                if executor is not None:
                    executor._shutdown = True
                # Notice other workers
                work_queue.put(None)
                return
            del executor
    except BaseException: # noqa
        _base.LOGGER.critical('Exception in worker', exc_info=True)


class ThreadPoolExecutor(BaseThreadPoolExecutor):

    def _adjust_thread_count(self):
        # if idle threads are available, don't spin new threads
        if self._idle_semaphore.acquire(timeout=0):
            return

        # When the executor gets lost, the weakref callback will wake up
        # the worker threads.
        def weakref_cb(_, q=self._work_queue):
            q.put(None) # noqa

        num_threads = len(self._threads)
        if num_threads < self._max_workers:
            thread_name = '%s_%d' % (self._thread_name_prefix or self,
                                     num_threads)
            t = threading.Thread(name=thread_name, target=_worker, daemon=True,
                                 args=(weakref.ref(self, weakref_cb),
                                       self._work_queue,
                                       self._initializer,
                                       self._initargs))
            t.start()
            self._threads.add(t) # noqa
            _threads_queues[t] = self._work_queue



from pathlib import Path


def scan_all_folders_with_rglob(root_path):
    """使用rglob获取所有文件夹"""
    root = Path(root_path)

    # 使用rglob('*')递归匹配所有路径，然后筛选出目录
    for p in root.rglob('*'):
        if not p.is_file():
            continue
        yield p




import requests

from .baai_config import DATACUBE_HOME, config_update

def login(ak, sk, login_api):

    save_name = DATACUBE_HOME / "config.json"

    headers = {
        "Accept-Language": "zh-CN",
        "Content-Type": "application/json",
    }
    resp = requests.post(login_api, json={"accessKey": ak, "secretKey": sk}, headers=headers)

    if resp.status_code != 200:
        print("登录失败")

    if resp.json().get("code") == 0:
        print(f"登录成功: config_path: {save_name.resolve()}")
    else:
        print("登录失败")

    token_data = {
        "access_key": ak,
        "secret_key": sk,
    }

    config_update(token_data)




import json

import requests

from .baai_config import DATACUBE_HOME, config_read

def download_meta(dataset_id: str, host="http://127.0.0.1:30201"):
    config_default = config_read()

    host = config_default.get("host", host)
    login_api = f"{host}/auth/user-access-key-login"
    config_path = DATACUBE_HOME / "config.json"

    with open(config_path) as f:
        config = json.load(f)

    access_key = config.get("access_key")
    secret_key = config.get("secret_key")

    resp_login = requests.post(login_api, json={"accessKey": access_key, "secretKey": secret_key})
    token = resp_login.json().get("data").get("token")

    meta_api = f"{host}/storage/download/{dataset_id}"
    headers = {
        "Accept-Language": "zh-CN",
        "Content-Type": "application/json",
    }
    req_header = {
        "Authorization": f"Bearer {token}",
    }
    headers.update(req_header)
    resp_meta = requests.get(meta_api, headers=headers)
    return resp_meta.text


def download_sign_meta(dataset_id: str, host="http://127.0.0.1:30201"):
    config_default = config_read() # noqa

    host = config_default.get("host", host)
    login_api = f"{host}/auth/user-access-key-login"
    config_path = DATACUBE_HOME / "config.json"

    with open(config_path) as f:
        config = json.load(f)

    access_key = config.get("access_key")
    secret_key = config.get("secret_key")

    resp_login = requests.post(login_api, json={"accessKey": access_key, "secretKey": secret_key})
    token = resp_login.json().get("data").get("token")

    meta_api = f"{host}/storage/download/sign/{dataset_id}"
    headers = {
        "Accept-Language": "zh-CN",
        "Content-Type": "application/json",
    }
    req_header = {
        "Authorization": f"Bearer {token}",
    }
    headers.update(req_header)
    resp_meta = requests.get(meta_api, headers=headers)
    return resp_meta.text


import requests

from .baai_config import config_update

def setup_network(cmd_args):
    try:
        private = "https://baai-datasets.ks3-cn-beijing-internal.ksyuncs.com/public/internal/speed/readme.md"
        requests.get(private, timeout=1)
    except requests.exceptions.ReadTimeout:
        network = "public"
    except Exception as e:
        print(e)
        network = "public"
    else:
        network = "private"

    config_data = {
        "network": network,
    }

    print("network: ", network)
    config_update(config_data)



import hashlib


CHUNK_SIZE = 1024 * 1024 * 5


def etag_bytes(bytes_):
    hash_md5 = hashlib.md5()
    hash_md5.update(bytes_)
    return hash_md5.digest()


def etag_write(path, size):

    part_etags = []

    for idx, start_pos in enumerate(range(0, size, CHUNK_SIZE)):
        end_pos = min(start_pos + CHUNK_SIZE, size)

        with path.open("rb") as reader:
            reader.seek(start_pos)
            rbytes = reader.read(end_pos - start_pos)
            part_etags.append(etag_bytes(rbytes))

    combined_md5 = b''.join(part_etags)
    final_md5_hex = hashlib.md5(combined_md5).hexdigest()

    return final_md5_hex


import os
import argparse
import pathlib
import traceback


from ..datacube_print import print_figlet
from ..baai_environment import setup_network


print_figlet()

datacube_home = pathlib.Path(os.path.expanduser("~")) / ".cache" / "datacube"
datacube_home.parent.mkdir(parents=True, exist_ok=True)


def runcmd_args(cmd_args=None):

    root_parser = argparse.ArgumentParser(add_help=False)
    root_parser.add_argument('--host', type=str, default="https://datacube.baai.ac.cn/api", help='服务地址')

    parser = argparse.ArgumentParser(prog='baai-datacube', description="数据魔方命令行工具")

    # 创建子命令解析器
    subparsers = parser.add_subparsers(dest='command')

    login_parser = subparsers.add_parser('login', help='登录数据魔方', parents=[root_parser])
    login_parser.set_defaults(func=login)

    down_parser = subparsers.add_parser('down', help='下载数据', parents=[root_parser])
    down_parser.set_defaults(func=down)
    down_parser.add_argument('--dataset', type=str, help='下载的数据集')
    down_parser.add_argument('-t', '--save-path', type=str, default=".", help='保存路径(默认当前文件夹)')
    down_parser.add_argument('--jobs', type=int, default=8, help='任务数')

    upload_parser = subparsers.add_parser('up', help='上传数据集', parents=[root_parser])
    upload_parser.add_argument('--collection', default=1,  type=int, help='下载的数据集')
    upload_parser.add_argument('-t', '--from-path', type=str, default=".", help='需要上传视频目录的文件')
    upload_parser.add_argument('--jobs', type=int, default=5, help='任务数')
    upload_parser.set_defaults(func=up)

    # 网络切换
    setup_network(cmd_args)

    # 解析命令行参数
    cmd_args = parser.parse_args()
    if hasattr(cmd_args, 'func'):
        try:
            cmd_args.func(cmd_args)
        except Exception: # noqa
            pass
        except KeyboardInterrupt:
            print()
            pass

    else:
        parser.print_help()


def login(cmd_args):
    from ..baai_login import login as login_user
    from ..baai_config import config_read

    config_default = config_read()
    host = config_default.get("host", cmd_args.host)

    print("---- login ----")

    ak = input(f"请输入ak [{config_default.get("access_key", "-")}]: ") or config_default.get("access_key")
    sk = input(f"请输入sk [{config_default.get("secret_key", "-")}]: ") or config_default.get("secret_key")
    login_api = f"{host}/auth/user-access-key-login"
    if not host.startswith("https"):
        print(f"login_api: {login_api}")
    login_user(ak, sk, login_api)


def down_old(cmd_args):
    from ..baai_config import Application

    config = Application(
        dataset_id=cmd_args.dataset,
        save_path=cmd_args.save_path,
        meta_path=pathlib.Path(cmd_args.save_path) / f"{cmd_args.dataset}.bin",
        host=cmd_args.host
    )

    from ..baai_meta import download_meta
    from ..baai_download import dataset_download


    config.try_login()

    save_path = pathlib.Path(cmd_args.save_path)
    save_path.mkdir(parents=True, exist_ok=True)
    print("---- download ----")

    dataset_meta = download_meta(cmd_args.dataset, host=cmd_args.host)
    meta_path = (save_path / f"{cmd_args.dataset}.bin")

    with open(meta_path, "w") as f:
        f.write(dataset_meta)

    dataset_download(
        dataset_id=cmd_args.dataset,
        save_path=save_path.resolve().__str__(),
        meta_path=meta_path.resolve().__str__(),
        jobs=cmd_args.jobs
    )


def down(cmd_args):
    try:

        from ..baai_config import Application

        config = Application(
            dataset_id=cmd_args.dataset,
            save_path=cmd_args.save_path,
            meta_path=pathlib.Path(cmd_args.save_path) / f"{cmd_args.dataset}.bin",
            host=cmd_args.host, jobs=cmd_args.jobs,
        )

        from ..baai_meta import download_sign_meta
        from ..baai_download_sign import download_sign_dataset

        try:
            config.try_login()
        except Exception as e:
            print(e)
            return

        save_path = pathlib.Path(cmd_args.save_path)
        save_path.mkdir(parents=True, exist_ok=True)
        print("---- download ----")

        dataset_meta = download_sign_meta(cmd_args.dataset, host=cmd_args.host)
        meta_path = (save_path / f"{cmd_args.dataset}.bin")

        with open(meta_path, "w") as f:
            f.write(dataset_meta)

        download_sign_dataset(
            dataset_id=cmd_args.dataset,
            save_path=save_path.resolve().__str__(),
            meta_path=meta_path.resolve().__str__(),
            jobs=cmd_args.jobs
        )
    except Exception: # noqa
        traceback.print_exc()


def up(cmd_args):

    print("\n---- upload ----\n")

    try:
        from ..baai_uploader import data_uploader
        from ..baai_config import Application

        config = Application(
            dataset_id=cmd_args.collection,
            save_path=cmd_args.from_path,
            meta_path=pathlib.Path(cmd_args.from_path) / f"{cmd_args.collection}.bin",
            host=cmd_args.host, jobs=cmd_args.jobs,
        )

        try:
            config.try_login()
        except Exception as e:
            print(e)
            return

        if not config.req_host.startswith("https"):
            print(f"host: {config.req_host}\n")

        data_uploader(cmd_args.collection, pathlib.Path(cmd_args.from_path), cmd_args.jobs)
    except Exception: # noqa
        traceback.print_exc()



import hashlib
import pathlib
import json
from concurrent.futures import ThreadPoolExecutor
from concurrent.futures import as_completed

import requests


from .baai_config import Application, Progress
from .serializer import MetaData


app = Application()
progress = Progress()

executor_spawn = ThreadPoolExecutor(max_workers=10)
executor_progress = ThreadPoolExecutor(max_workers=app.jobs)
executor_sign = ThreadPoolExecutor(max_workers=app.jobs)
executor_down = ThreadPoolExecutor(max_workers=app.jobs)
executor_watch = ThreadPoolExecutor(max_workers=50)
executor_merge = ThreadPoolExecutor(max_workers=app.jobs/2)



def download_readrows():
    config = Application()
    with open(config.mate_path, "r") as fr:
        for _line in fr:
            progress.add_require_count(1)

def check_file(file_path: str, length: int):
    path = pathlib.Path(file_path)
    if not path.exists():
        return False
    check_size = path.stat().st_size
    return check_size== length


def download_sign(file_path: str):
    config = Application()
    req_headers = {
        "Authorization": f"Bearer {config.try_login()}",
    }
    req_headers.update(config.req_header)
    resp_sign = requests.post(
        config.sign_api, headers=req_headers, json={"paths": [file_path], "network": config.network}
    )
    sign_addr = resp_sign.json().get("data")[0].get("endpoint")
    return sign_addr


def download_file(raw_data: str):
    config = Application()

    meta_file = MetaData(**json.loads(raw_data))

    hash_prefix = hashlib.md5(meta_file.get_storage_path().encode("utf-8")).hexdigest()
    hash_ext = pathlib.Path(meta_file.get_storage_path()).suffix

    rs_sign = executor_sign.submit(download_sign, meta_file.get_storage_path())
    sign_addr:str = rs_sign.result()
    resp_meta = requests.head(sign_addr, headers={"accept-encoding": None})
    size = int(resp_meta.headers.get('Content-Length'))

    progress.add_require_size(size)
    # 文件存在不进行下载
    if check_file(f"{config.save_path}/{hash_prefix}{hash_ext}", size):
        progress.add_download_size(size)
        progress.add_download_count(1)
        return

    chunk_size = config.chunk_size
    total_parts = int((size + chunk_size - 1) / chunk_size)

    save_path = pathlib.Path(config.save_path)
    save_path.mkdir(parents=True, exist_ok=True)

    download_futures = []
    for part_index in range(total_parts):
        start = part_index * chunk_size
        end = min(start + chunk_size, size)
        task_down = executor_down.submit(download_part_write, meta_file, start, end, hash_prefix)
        download_futures.append(task_down)

    # 等待完成
    for future in as_completed(download_futures):
        future.result()

    executor_merge.submit(download_part_merge,  hash_prefix, hash_ext, total_parts)

def download_part_write(meta_file: MetaData, start: int, end: int, part_hash):
    config = Application()

    part_idx = int(start / config.chunk_size)
    part_name =  pathlib.Path(f"{config.save_path}/{part_hash}_{part_idx}.bin")

    if check_file(part_name.__str__(),  end-start):
        pass
    else:
        range_header = f"bytes={start}-{end - 1}"
        sign_addr = download_sign(meta_file.get_storage_path())
        resp_file = requests.get(sign_addr, headers={"Range": range_header})

        # md5加密
        with part_name.open("wb") as fwriter:
            fwriter.write(resp_file.content)

    progress.add_download_size(end-start)

def download_part_merge(name_prefix, name_ext, total_parts):
    import mmap

    config = Application()
    target_path = pathlib.Path(f"{config.save_path}/{name_prefix}{name_ext}")
    with target_path.open("ab") as fwriter:
        for part_idx in range(total_parts):
            part_name = pathlib.Path(f"{config.save_path}/{name_prefix}_{part_idx}.bin")
            with part_name.open("rb") as f:
                with mmap.mmap(f.fileno(), length=0, access=mmap.ACCESS_READ) as mm:
                    fwriter.write(mm)

    for part_idx in range(total_parts):
        part_name = pathlib.Path(f"{config.save_path}/{name_prefix}_{part_idx}.bin")
        part_name.unlink()
    progress.add_download_count(1)


def download_progress():
    config = Application()
    with open(config.mate_path, "r") as fr:
        for _line in fr:
            try:
                executor_progress.submit(download_file, _line)
            except Exception as e: # noqa
                progress.add_download_count(-1)

def download_watch():
    import time

    while True:
        time.sleep(0.1)
        if progress.require_count == progress.download_count:
            break
    return progress.require_count







import datetime
import hashlib
import msgpack
import base64
import jwt


def jwtsign_create(proto, prefix, path, check_point, secret="secret"):
    """jwtsign_create

    创建用户需要上传云存储的数据签名
    """

    check_point = hashlib.md5(check_point.encode("utf-8")).hexdigest()

    upload_pack = msgpack.dumps(
        {"proto":proto, "prefix": prefix, "path": path},
    )
    upload_path = base64.b64encode(upload_pack).decode("utf-8")

    payload = {
        "upload_path": upload_path,
        "check_point": check_point,
        "stamp": int(datetime.datetime.now().timestamp()),
    }
    download_sign = jwt.encode(payload, secret, algorithm="HS256")


    return download_sign


if __name__ == "__main__":

    import  requests

    sign_data = jwtsign_create("ks3://baai-data-label", "202511261801", "test.txt", "127.0.0.1")


    resp_multipart = requests.post(
        "http://127.0.0.1:31000/v1/storage/upload/multipart",
        json={
            "upload_sign": sign_data,
        }
    )

    print(resp_multipart.status_code)
    print(resp_multipart.text)
    print("============="* 100)

    upload_id = resp_multipart.json().get("data").get("upload_id")
    resp = requests.post(
        "http://127.0.0.1:31000/v1/storage/upload/presign",
        json={
            "upload_sign": sign_data,
            "upload_id": upload_id,
            "upload_num": 1,
            "network": "public"
        }
    )

    print(resp.status_code)
    print(resp.text)

    print("============="* 100)

    with open("/Users/hgshicc/Downloads/s3_paths.txt", "rb") as f:
        resp_write = resp_upload = requests.put(
            resp.json().get("data").get("endpoint"),
            data=f
        )

        print(resp_write.status_code)
        for k, v in resp_write.headers.items():
            print(k, v)

    e_tag = resp_write.headers.get("ETag")

    print("============="* 100)
    resp_complete = resp_multipart = requests.post(
        "http://127.0.0.1:31000/v1/storage/upload/multipart",
        json={
            "upload_sign": sign_data,
            "upload_id": upload_id,
            "upload_parts": [e_tag]
        }
    )
    print(resp_complete.status_code)





import cv2


def get_video_duration(video_path):
    """
    获取视频文件的时长（秒）
    返回值: 成功返回时长(秒)，失败返回None
    """
    cap = cv2.VideoCapture(video_path)

    if not cap.isOpened():
        print(f"错误：无法打开视频文件 {video_path}")
        return None

    # 获取关键属性
    total_frames = int(cap.get(cv2.CAP_PROP_FRAME_COUNT))
    fps = cap.get(cv2.CAP_PROP_FPS)

    # 释放视频流
    cap.release()

    # 检查并计算时长
    if fps > 0 and total_frames > 0:
        duration_seconds = total_frames / fps
        return duration_seconds
    else:
        print("警告：无法获取有效的帧数或帧率。")
        return None



import time
import traceback
import functools
import logging


Logger = logging.getLogger(__name__)


def retry(max_retries=3, retry_delay=1):
    """
    重试装饰器

    Args:
        max_retries: 最大重试次数
        retry_delay: 重试间隔时间(秒)
    """

    def decorator(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            last_exception = None
            func_name = func.__name__

            for idx, attempt in enumerate(range(max_retries + 1)):
                try:
                    return func(*args, **kwargs)
                except Exception as e:
                    last_exception = e
                    if attempt < max_retries:
                        # Logger.warning(f"{func_name}, 第{attempt + 1}次调用失败: {e}")
                        time.sleep(retry_delay * 2 * (idx+1))
                    else:
                        traceback.print_exc()
                        Logger.warning(f"{func_name}, 所有重试均失败，共尝试{max_retries + 1}次: {e}")
            # 抛出最后一次异常
            raise last_exception
        return wrapper

    return decorator




import traceback
from concurrent.futures import wait, FIRST_COMPLETED


import requests

from tqdm import tqdm

from ..baai_config import Application
from . import baai_presign, baai_retry


## 上传块大小
CHUNK_SIZE = 1024 * 1024 * 5

@baai_retry.retry(max_retries=3)
def sign_write(collection_id, file_name, file_size, file_etag, video_duration):
    """sign_write

    文件上传签名，如果文件上传了一部分，使用返回的数据进行断点续传
    """

    app = Application()
    req_headers = {
        "Authorization": f"Bearer {app.try_login()}",
    }

    req_host = app.req_host

    try:
        resp_sign = requests.post(
            f"{req_host}/collection/file/create",
            headers=req_headers,
            json={
                "fileEtag": file_etag,
                "collectionId": collection_id,
                "fileName": file_name,
                "fileSize": file_size,
                "videoDuration": video_duration
            }, timeout=(1, 5)
        )

    except Exception as e:
        raise e
    else:
        assert resp_sign.status_code == 200 and resp_sign.json().get("code") == 0, resp_sign.text

        resp_data = resp_sign.json().get("data")
        sign_data = resp_data.get("uploadSign")

        return sign_data, resp_data.get("uploadId"), resp_data.get("uploadSize")


@baai_retry.retry(max_retries=3)
def update_write(collection_id, file_etag, file_size, upload_id=""):
    """update_write

    文件上传进度记录
    """


    app = Application()
    req_headers = {
        "Authorization": f"Bearer {app.try_login()}",
    }


    req_host = app.req_host

    resp_sign = requests.post(
        f"{req_host}/collection/file/update",
        headers=req_headers,
        json={
            "collectionId": collection_id,
            "fileEtag": file_etag,
            "uploadSize": file_size,
            "uploadId": upload_id
        }
    )

    assert resp_sign.status_code == 200 and resp_sign.json().get("code") == 0, resp_sign.text


def multipart_upload(
        collection_id, etag, size, video_path, sign_data,
        upload_id, upload_parts,
        video_number, video_name,
        executor,
):

    chunk_tags =  []
    m_parts = {}
    if upload_parts:
        m_parts = { part.get("part_number")-1: part.get("e_tag") for part in upload_parts}

    total_length = 0
    tasks = []

    if len(video_name) > 18:
        desc_name = "..." + video_name[-15:]
    else:
        desc_name = video_name.rjust(18)

    bar_format= " {desc:>20} | {elapsed:>10} | {percentage:>5.1f}%|{bar}| {n_fmt:>8}/{total_fmt:<8}{postfix} | {rate_fmt:>10} | ETA {remaining:>8}"
    with tqdm(ncols=120, ascii="░▒▓█", total=size, unit='B', unit_scale=True, desc=desc_name, position=video_number, leave=False, bar_format=bar_format) as pbar:
        for idx, start_pos in enumerate(range(0, size, CHUNK_SIZE)):
            end_pos = min(start_pos + CHUNK_SIZE, size)

            # 跳过已经上传的分片
            if e_tag := m_parts.get(idx+1):
                chunk_tags.append(e_tag)

                total_length += end_pos - start_pos
                update_write(collection_id, etag, end_pos, upload_id)

                pbar.update(end_pos - start_pos)
                continue

            # 读取分片数据,进行上传
            with video_path.open("rb") as reader:
                reader.seek(start_pos)
                write_bytes = reader.read(end_pos - start_pos)

                try:
                    task_write = executor.submit(baai_presign.write_multipart_data, write_bytes, sign_data, upload_id, idx)
                    tasks.append(task_write)
                except Exception as e:
                    raise e

        pending = set(tasks)

        while pending:
            done, pending = wait(pending, timeout=0.5, return_when=FIRST_COMPLETED)
            for task in done:
                try:
                    chunk_etag, length = task.result()
                    chunk_tags.append(chunk_etag)
                    total_length += length
                except Exception as e:
                    traceback.print_exc()
                    print(e)
                else:
                    # 更新进度
                    update_write(collection_id, etag, total_length, upload_id)
                    pbar.update(length)

    return chunk_tags



import time
import os
import pathlib
import threading
import json

import jwt
import requests


DATACUBE_HOME = pathlib.Path(os.path.expanduser("~")) / ".cache" / "datacube"
DATACUBE_HOME.parent.mkdir(parents=True, exist_ok=True)
HOME_CONFIG = DATACUBE_HOME / "config.json"

REQUEST_HEADER = {
    "Accept-Language": "zh-CN",
    "Content-Type": "application/json",
}

class Application:
    _initialized: bool = False

    def __new__(cls, *args, **kwargs) -> "Application":
        if not hasattr(cls, "_instance"):
            cls._instance = super().__new__(cls)
        return cls._instance # noqa

    def __init__(self, dataset_id: str="",  save_path=".", meta_path="", jobs=8, host: str = "http://datacube.baai.ac.cn/api"):
        if not self._initialized:

            self.root_path = DATACUBE_HOME
            self.req_header = REQUEST_HEADER
            self.req_host =  host
            config_default = config_read()
            host_default = config_default.get("host")
            if host_default:
                self.req_host = host_default
                host = host_default
            self.network = config_default.get("network", "public")

            self.sign_api = f"{host}/v1/datacube/storage/sign-generate"
            self.login_api = f"{host}/v1/datacube/auth/user-access-key-login"
            self.logger_api = f"{host}/v1/datacube/logger/download/create"

            self.dataset_id = dataset_id # 下载的数据集
            self.save_path = save_path
            self.mate_path = meta_path
            self.chunk_size = 1024 * 1024 * 5
            self.jobs = jobs

            self._login_token = ""
            self._login_lock = threading.Lock()

            self._initialized = True

    def try_access_token(self):
        save_name = self.root_path / "config.json"

        with open(save_name, "r") as fr:
            data = json.loads(fr.read())
        resp_login = requests.post(self.login_api, json={"accessKey": data.get("access_key"), "secretKey": data.get("secret_key")}, headers=self.req_header)

        resp_data = resp_login.json()

        assert resp_data.get("code") == 0, resp_data.get("message")
        return resp_data.get("data").get("token")

    def update_token(self):
        with self._login_lock:
            if self._login_token == "":
                self._login_token = self.try_access_token()

    def try_login(self):

        if self._login_token == "":
            self.update_token()

        decoded = jwt.decode(self._login_token, options={"verify_signature": False})
        exp = decoded.get("exp")
        if exp + 60 * 10 < int(time.time()):
            self.update_token()
        return self._login_token


class Progress:
    _initialized: bool = False

    def __new__(cls, *args, **kwargs) -> "Application":
        if not hasattr(cls, "_instance"):
            cls._instance = super().__new__(cls)
        return cls._instance # noqa

    def __init__(self):
        if not self._initialized:
            self._lock = threading.Lock()
            self.require_count = 0 # 需要下载的数量
            self.require_size = 0 # 需要下载的大小

            self.download_count = 0 # 下载数量
            self.download_size = 0 # 下载大小

            # 失败数量
            self.fail_count = 0

            self._initialized = True

    def add_require_count(self, count):
        with self._lock:
            self.require_count += count

    def add_require_size(self, size):
        with self._lock:
            self.require_size += size

    def add_download_count(self, count):
        with self._lock:
            self.download_count += count

    def add_download_size(self, size):
        with self._lock:
            self.download_size += size

    def add_fail_count(self, count):
        with self._lock:
            self.fail_count += count

def config_update(config_data):
    try:
        with open(HOME_CONFIG, 'r') as f:
            config_default = json.load(f)
    except (FileNotFoundError, ValueError, json.JSONDecodeError):
        config_default = {}
    config_default.update(config_data)

    try:
        HOME_CONFIG.parent.mkdir(parents=True, exist_ok=True)
        with open(HOME_CONFIG, 'w') as f:
            json.dump(config_default, f, indent=4)
    except Exception as e:
        print(f"保存配置文件时出错: {e}")

def config_read():
    try:
        with open(HOME_CONFIG, 'r') as f:
            config_default = json.load(f)
    except (FileNotFoundError, ValueError, json.JSONDecodeError):
        config_default = {}
    return config_default

def print_config():
    try:
        with open(HOME_CONFIG, 'r') as f:
            config_default = json.load(f)
    except (FileNotFoundError, ValueError, json.JSONDecodeError):
        config_default = {}

    for k, v in config_default.items():
        if k == "sk":
            continue
        print(f"{k}: {v}")
    print("")



import requests


from pyrate_limiter import Rate, Duration, Limiter

from ..baai_config import Application

from .baai_retry import retry


# 存储相关的服务地址
# STORAGE = "http://127.0.0.1:31000"
# STORAGE = "http://120.92.15.143/api"

# 控制签名速度
LIMITER = Limiter(Rate(20, Duration.SECOND * 1), max_delay=10000)


@retry(max_retries=3)
def write_presign(sign_data, upload_id, idx):
    config = Application()

    post_data = {
        "upload_sign": sign_data,
        "upload_id": upload_id,
        "upload_num": idx + 1,
        "network": config.network
    }

    host = config.req_host

    sign_api = f"{host}/v1/storage/upload/presign"
    resp_sign = requests.post(sign_api, json=post_data)

    assert resp_sign.status_code == 200 and resp_sign.json().get("code") == 0, resp_sign.text

    # 成功限制速度
    LIMITER.try_acquire("write_presign")

    return resp_sign.json().get("data").get("endpoint")


@retry(max_retries=3)
def read_presign(sign_data):
    config = Application()

    post_data = {
        "download_sign": sign_data.strip(),
        "network": "public"
    }

    host = config.req_host

    sign_api = f"{host}/v1/storage/download/presign"
    # sign_api = f"http://127.0.0.1:8085/v1/storage/download/presign"
    resp_sign = requests.post(sign_api, json=post_data)

    assert resp_sign.status_code == 200 and resp_sign.json().get("code") == 0, resp_sign.text

    # # 成功限制速度
    # LIMITER.try_acquire("read_presign")

    endpoint = resp_sign.json().get("data").get("endpoint")
    return endpoint


@retry(max_retries=3)
def write_multipart(sign_data, upload_id=None, etags=None):
    """write multipart

    分片上传的开始与结束(when upload_id not none and all etags)

    场景:
    1. upload_id: none, 创建 upload_id
    2. upload_id: v, etags: none, 获取以及上传的分片信息
    3. upload_id: v, etags: v, 合并分片完成上传
    """


    config = Application()
    post_data = {"upload_sign": sign_data}

    if upload_id:
        post_data.update({"upload_id": upload_id})

    if etags:
        post_data.update({"upload_parts": etags})

    host = config.req_host
    multipart_api = f"{host}/v1/storage/upload/multipart"
    resp_multipart = requests.post(multipart_api, json=post_data, timeout=5)

    assert resp_multipart.status_code == 200, resp_multipart.text
    resp_data = resp_multipart.json().get("data")
    upload_id = resp_data.get("upload_id")
    upload_parts = resp_data.get("upload_parts")
    return upload_id, upload_parts


@retry(max_retries=10)
def write_multipart_data(
    write_bytes, sign_data, upload_id, idx,
):
    """write_multipart_data

    上传分片数据
    """

    # presign
    write_endpoint = write_presign(sign_data, upload_id, idx)

    # write
    resp_write = requests.put(write_endpoint, data=write_bytes)
    e_tag = resp_write.headers.get("ETag")

    # 判断是否上传成功
    assert resp_write.status_code == 200

    return e_tag, len(write_bytes)



import os
import time
import logging
import traceback
from concurrent.futures import wait, FIRST_COMPLETED

import requests
from tqdm import tqdm

from .core import baai_write, baai_etag, baai_presign, baai_concurrent
from .helper import baai_vedio, baai_scanner
from .baai_config import Application


Logger = logging.getLogger(__name__)


def data_uploader(collection, upload_path=".", job_workers=5):
    # 环境变量
    collection_id = collection
    try:

        video_count = sum(1 for _ in baai_scanner.scan_all_folders_with_rglob(upload_path))

        # 线程池环境
        start = time.time()
        tasks = []

        worker_pool = baai_concurrent.ThreadPoolExecutor(max_workers=job_workers)
        writer_pool = baai_concurrent.ThreadPoolExecutor(max_workers=job_workers*2)

        bar_format= "[{elapsed}] | {percentage:5.1f}%|{bar}| {n_fmt}/{total_fmt}{postfix}"
        with tqdm(ncols=120, ascii="░▒▓█", total=video_count, desc="", position=0, leave=True, bar_format=bar_format) as pbar:
            for video_number, video_path in enumerate(baai_scanner.scan_all_folders_with_rglob(upload_path)):
                try:
                    pbar_idx = video_number % 5 + 1
                    task_worker = worker_pool.submit(
                        multipart_uploader,
                        video_path, pbar_idx+1, collection_id, writer_pool
                    )
                    tasks.append(task_worker)
                except Exception as e:
                    traceback.print_exc()
                    Logger.error(e)

            completed_count = 0
            pending = set(tasks)  # 所有 futures

            try:
                while pending:
                    done, pending = wait(pending, timeout=0.5, return_when=FIRST_COMPLETED)
                    for task in done:
                        try:
                            success_count = task.result()
                            completed_count += success_count
                        except Exception as e:
                            traceback.print_exc()
                            Logger.error(e)
                        finally:
                            pbar.update(1)
            except KeyboardInterrupt:
                # writer_pool.shutdown(wait=False, cancel_futures=True)
                # worker_pool.shutdown(wait=False, cancel_futures=True)
                raise

        end = time.time()
        print()
        print(f"Success: {completed_count}, Fail: {video_count - completed_count}, Total {video_count}")
        print()
        print(f"{end - start:.2f}s")

    except KeyboardInterrupt:

        app = Application()
        req_headers = {
            "Authorization": f"Bearer {app.try_login()}",
        }

        host = app.req_host

        # TODO: 增加重试
        requests.put(
            f"{host}/storage/upload/{collection_id}",
            headers=req_headers,
            json={
                "status": "cancelled",
            }
        )


        print("\n\n\n\n\n\n\n\n\n\n\n")
        print("=================已经取消上传==================")
        os._exit(1)

    except Exception as e:
        traceback.print_exc()
        print(e)
    else:

        app = Application()
        req_headers = {
            "Authorization": f"Bearer {app.try_login()}",
        }

        host = app.req_host

        requests.put(
            f"{host}/storage/upload/{collection_id}",
            headers=req_headers,
            json={
                "status": "completed",
            }
        )

    finally:
        pass


def multipart_uploader(video_path, video_number, collection_id, writer_executor):
    try:

        # 获取视频时长
        try:
            video_duration = int(baai_vedio.get_video_duration(video_path))
        except Exception as e:
            print(video_path.name, e)
            return 0

        size = video_path.stat().st_size
        etag = baai_etag.etag_write(video_path, size)


        # 创建上传的文件信息
        sign_data, upload_id, upload_size = baai_write.sign_write(collection_id, f"{video_path.name}", size, etag, video_duration)

        # 检查是否上传完成
        if upload_size == size:
            return 1
        else:
            upload_id, upload_parts = baai_presign.write_multipart(sign_data, upload_id=upload_id)

        if not upload_id:
            upload_id, upload_parts = baai_presign.write_multipart(sign_data)

        if not upload_id:
            return 0

        # 上传所有分片
        e_tags = baai_write.multipart_upload(
            collection_id,  # 数据集信息
            etag, size,  # 文件信息
            video_path, # 文件路径
            sign_data,  # 签名数据
            upload_id,  # 上传信息
            upload_parts, # 上传的分片信息
            video_number, # 分片ID
            video_path.name, # 文件名字
            writer_executor # 写入线程池
        )

        # 完成分片上传
        baai_presign.write_multipart(sign_data, upload_id, e_tags)
    except Exception: # noqa
        traceback.print_exc()
        return 0
    else:
        # 成功增加完成量
        return 1



import uuid
import requests


process_id = uuid.uuid4().__str__().replace("-", "")[:32]

def logger_download():
    from .baai_config import  Application, Progress

    app = Application()
    app.try_login()

    headers = {
        "Authorization": f"Bearer {app.try_login()}",
    }
    progress = Progress()
    headers.update(app.req_header)
    post_data = {"datasetId": app.dataset_id, "progressId": process_id, "downloadSize": progress.download_size}

    _resp = requests.post(app.logger_api, json=post_data,headers=headers)
